package dev.fitko.fitconnect.core.validation;

import com.fasterxml.jackson.core.JacksonException;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.networknt.schema.JsonSchemaFactory;
import com.networknt.schema.SpecVersion;
import com.networknt.schema.ValidationMessage;
import com.nimbusds.jose.jwk.KeyOperation;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.util.Base64;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.model.event.EventClaimFields;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.DataIntegrityException;
import dev.fitko.fitconnect.api.exceptions.SchemaNotFoundException;
import dev.fitko.fitconnect.api.exceptions.ValidationException;
import dev.fitko.fitconnect.api.services.crypto.MessageDigestService;
import dev.fitko.fitconnect.api.services.schema.SchemaProvider;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import dev.fitko.fitconnect.core.util.Strings;
import dev.fitko.fitconnect.jwkvalidator.JWKValidator;
import dev.fitko.fitconnect.jwkvalidator.JWKValidatorBuilder;
import dev.fitko.fitconnect.jwkvalidator.exceptions.JWKValidationException;
import dev.fitko.fitconnect.jwkvalidator.exceptions.LogLevel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.Proxy;
import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import static dev.fitko.fitconnect.jwkvalidator.JWKValidator.withX5CValidation;
import static dev.fitko.fitconnect.jwkvalidator.JWKValidator.withoutX5CValidation;

public class DefaultValidationService implements ValidationService {

    private static final Logger LOGGER = LoggerFactory.getLogger(DefaultValidationService.class);
    private static final ObjectMapper MAPPER = new ObjectMapper().setDateFormat(new SimpleDateFormat("yyyy-MM-dd"));
    private static final JsonSchemaFactory SCHEMA_FACTORY = JsonSchemaFactory.getInstance(SpecVersion.VersionFlag.V202012);
    public static final String VALID_SCHEMA_URL_EXPRESSION = "https://schema\\.fitko\\.de/fit-connect/metadata/1\\.\\d+\\.\\d+/metadata.schema.json";

    private final MessageDigestService messageDigestService;
    private final SchemaProvider schemaProvider;
    private final ApplicationConfig config;

    public DefaultValidationService(final ApplicationConfig config, final MessageDigestService messageDigestService, final SchemaProvider schemaProvider) {
        this.config = config;
        this.messageDigestService = messageDigestService;
        this.schemaProvider = schemaProvider;
    }

    @Override
    public ValidationResult validateEncryptionPublicKey(final RSAKey publicKey) {
        try {
            validateKey(publicKey, KeyOperation.WRAP_KEY);
        } catch (final Exception e) {
            return ValidationResult.error(e);
        }
        return ValidationResult.ok();
    }

    @Override
    public ValidationResult validateSignaturePublicKey(final RSAKey signatureKey) {
        try {
            validateKey(signatureKey, KeyOperation.VERIFY);
        } catch (final Exception e) {
            return ValidationResult.error(e);
        }
        return ValidationResult.ok();
    }

    @Override
    public ValidationResult validateSetEventSchema(final String setEventPayload) {
        try {
            final JsonNode inputNode = MAPPER.readTree(setEventPayload);
            final URI schemaUri = URI.create(inputNode.get(EventClaimFields.CLAIM_SCHEMA).asText());
            if (schemaProvider.isAllowedSetSchema(schemaUri)) {
                return validateJsonSchema(schemaProvider.loadLatestSetSchema(), inputNode);
            } else {
                return ValidationResult.error(new SchemaNotFoundException("SET payload schema not supported: " + schemaUri));
            }
        } catch (final JsonProcessingException | IllegalArgumentException e) {
            return ValidationResult.error(e);
        }
    }

    @Override
    public ValidationResult validateMetadataSchema(final Metadata metadata) {

        if (metadata.getSchema() == null || metadata.getSchema().isEmpty()) {
            return ValidationResult.ok();
        } else if (!metadata.getSchema().matches(VALID_SCHEMA_URL_EXPRESSION)) {
            return ValidationResult.error(new ValidationException("The provided metadata schema is not supported."));
        }

        try {
            final String metadataJson = MAPPER.writeValueAsString(metadata);
            final JsonNode inputNode = MAPPER.readTree(metadataJson);
            return validateJsonSchema(schemaProvider.loadMetadataSchema(config.getMetadataSchemaWriteVersion()), inputNode);
        } catch (final JsonProcessingException e) {
            return ValidationResult.error(e);
        }
    }

    @Override
    public ValidationResult validateHashIntegrity(final String originalHexHash, final byte[] data) {
        try {
            final byte[] originalHash = messageDigestService.fromHexString(originalHexHash);
            final boolean hashesAreNotEqual = !messageDigestService.verify(originalHash, data);
            if (hashesAreNotEqual) {
                return ValidationResult.error(new DataIntegrityException("Hash sum of transmitted data does not equal the hash of the sender."));
            }
            return ValidationResult.ok();
        } catch (final IllegalArgumentException | NullPointerException e) {
            return ValidationResult.error(e);
        }
    }

    @Override
    public ValidationResult validateJsonFormat(final String json) {
        try {
            MAPPER.readTree(json);
        } catch (final JacksonException e) {
            return ValidationResult.error(e);
        }
        return ValidationResult.ok();
    }

    @Override
    public ValidationResult validateXmlFormat(final String xml) {
        try {
            final XMLReader xmlReader = getXmlReader();
            xmlReader.parse(new InputSource(new ByteArrayInputStream(xml.getBytes(StandardCharsets.UTF_8))));
        } catch (final ParserConfigurationException | IOException | SAXException e) {
            return ValidationResult.error(e);
        }
        return ValidationResult.ok();
    }

    private ValidationResult validateJsonSchema(final String schema, final JsonNode inputNode) {
        return returnValidationResult(SCHEMA_FACTORY.getSchema(schema).validate(inputNode));
    }

    private void validateKey(final RSAKey publicKey, final KeyOperation purpose) throws JWKValidationException, CertificateEncodingException {
        if (config.getCurrentEnvironment().isAllowInsecurePublicKey()) {
            validateWithoutCertChain(publicKey, purpose);
        } else {
            validateCertChain(publicKey, purpose);
        }
    }

    private ValidationResult returnValidationResult(final Set<ValidationMessage> errors) {
        if (errors.isEmpty()) {
            return ValidationResult.ok();
        }
        return ValidationResult.error(new ValidationException(errorsToSingleString(errors)));
    }

    private void validateWithX509AndProxy(final List<String> pemCerts, final RSAKey publicKey, final KeyOperation purpose) throws JWKValidationException {
        final var proxy = new Proxy(Proxy.Type.HTTP, new InetSocketAddress(config.getHttpProxyHost(), config.getHttpProxyPort()));
        LOGGER.info("Using proxy {} for key validation", proxy);
        addLogLevel(withX5CValidation()
                .withProxy(proxy)
                .withRootCertificatesAsPEM(pemCerts))
                .validate(publicKey, purpose);
    }


    private void validateWithX509AndWithoutProxy(final List<String> pemCerts, final RSAKey publicKey, final KeyOperation purpose) throws JWKValidationException {
        LOGGER.info("No proxy set for validation");
        addLogLevel(withX5CValidation()
                .withoutProxy()
                .withRootCertificatesAsPEM(pemCerts))
                .validate(publicKey, purpose);
    }

    private JWKValidator addLogLevel(final JWKValidatorBuilder.JWKValidatorX5CErrorHandling validator) {
        if (validateSilent()) {
            return validator.withoutThrowingExceptions().withErrorLogLevel(LogLevel.WARN).build();
        } else {
            return validator.withThrowingExceptions().withErrorLogLevel(LogLevel.ERROR).build();
        }
    }

    private void validateWithoutCertChain(final RSAKey publicKey, final KeyOperation purpose) throws JWKValidationException {
        LOGGER.info("Validating public key without XC5 certificate chain");
        withoutX5CValidation().withErrorLogLevel(validateSilent() ? LogLevel.WARN : LogLevel.ERROR).build().validate(publicKey, purpose);
    }

    private void validateCertChain(final RSAKey publicKey, final KeyOperation purpose) throws JWKValidationException, CertificateEncodingException {
        final List<String> pemCerts = getPemCerts(publicKey.getParsedX509CertChain());
        LOGGER.info("Validation public key with XC5 certificate chain checks");
        if (isProxySet()) {
            validateWithX509AndProxy(pemCerts, publicKey, purpose);
        } else {
            validateWithX509AndWithoutProxy(pemCerts, publicKey, purpose);
        }
    }

    private List<String> getPemCerts(final List<X509Certificate> certChain) throws CertificateEncodingException {
        final List<String> pemCerts = new ArrayList<>();
        for (final X509Certificate cert : certChain) {
            pemCerts.add(getEncodedString(cert));
        }
        return pemCerts;
    }

    private String getEncodedString(final X509Certificate x509Certificate) throws CertificateEncodingException {
        return Base64.encode(x509Certificate.getEncoded()).toString();
    }

    private boolean isProxySet() {
        return config.getHttpProxyPort() != null && Strings.isNotNullOrEmpty(config.getHttpProxyHost());
    }
    
    private boolean validateSilent() {
        return config.getCurrentEnvironment().isAllowInsecurePublicKey();
    }

    private String errorsToSingleString(final Set<ValidationMessage> errors) {
        return errors.stream()
                .map(ValidationMessage::getMessage)
                .collect(Collectors.joining("\n"));
    }

    private static XMLReader getXmlReader() throws ParserConfigurationException, SAXException {
        final SAXParserFactory saxParserFactory = SAXParserFactory.newInstance();
        saxParserFactory.setFeature("http://apache.org/xml/features/disallow-doctype-decl", true);
        return saxParserFactory.newSAXParser().getXMLReader();
    }

}
